use std::collections::HashMap;

use lazy_static::lazy_static;

use super::item::ItemId;

#[derive(Debug, thiserror::Error)]
#[error("{0} is not a dungeon room")]
pub struct NotADungeonRoom(u8);

#[derive(Clone, Copy, Eq, Hash, PartialEq)]
pub enum Dungeon {
    HyruleCastle,
    EasternPalace,
    DesertPalace,
    TowerOfHera,
    AgahnimsTower,
    PalaceOfDarkness,
    SwampPalace,
    SkullWoods,
    ThievesTown,
    IcePalace,
    MiseryMire,
    TurtleRock,
    GanonsTower,
}

pub fn get_dungeon_from_room(room_id: u8) -> Result<Dungeon, NotADungeonRoom> {
    for (dungeon, rooms) in DUNGEON_ROOMS.iter() {
        if let Ok(_) = rooms.binary_search(&room_id) {
            return Ok(*dungeon);
        }
    }

    Err(NotADungeonRoom(room_id))
}

const HYRULE_CASTLE_ROOMS: [u8; 21] = [
    1, 2, 17, 18, 33, 34, 50, 65, 66, 80, 81, 82, 96, 97, 98, 112, 113, 114, 128, 129, 130,
];

const EASTERN_PALACE_ROOMS: [u8; 13] = [
    137, 153, 168, 169, 170, 184, 185, 186, 200, 201, 216, 217, 218,
];

const DESERT_PALACE_ROOMS: [u8; 10] = [51, 67, 83, 99, 115, 116, 117, 131, 132, 133];

const TOWER_OF_HERA_ROOMS: [u8; 7] = [7, 23, 39, 49, 119, 135, 167];

const AGAHNIMS_TOWER_ROOMS: [u8; 7] = [32, 48, 64, 176, 192, 208, 224];

const PALACE_OF_DARKNESS_ROOMS: [u8; 14] = [9, 10, 11, 25, 26, 27, 42, 43, 58, 59, 74, 75, 90, 106];

const SWAMP_PALACE_ROOMS: [u8; 13] = [6, 22, 38, 40, 52, 53, 54, 55, 56, 70, 84, 102, 118];

const SKULL_WOODS_ROOMS: [u8; 9] = [41, 57, 73, 86, 87, 88, 89, 103, 104];

const THIEVES_TOWN_ROOMS: [u8; 12] = [68, 69, 100, 101, 171, 172, 187, 188, 203, 204, 219, 220];

const ICE_PALACE_ROOMS: [u8; 22] = [
    14, 30, 31, 46, 62, 63, 78, 79, 94, 95, 110, 126, 127, 142, 158, 159, 174, 175, 190, 191, 206,
    222,
];

const MISERY_MIRE_ROOMS: [u8; 18] = [
    144, 145, 146, 147, 151, 152, 160, 161, 162, 163, 177, 178, 179, 193, 194, 195, 209, 210,
];

const TURTLE_ROCK_ROOMS: [u8; 17] = [
    4, 19, 20, 21, 35, 36, 164, 180, 181, 182, 183, 196, 197, 198, 199, 213, 214,
];

const GANONS_TOWER_ROOMS: [u8; 26] = [
    12, 13, 28, 29, 61, 76, 77, 91, 92, 93, 107, 108, 109, 123, 124, 125, 139, 140, 141, 149, 150,
    155, 156, 157, 165, 166,
];

lazy_static! {
    static ref DUNGEON_ROOMS: HashMap<Dungeon, &'static [u8]> = {
        use Dungeon::*;

        let mut rooms = HashMap::new();

        rooms.insert(HyruleCastle, &HYRULE_CASTLE_ROOMS[..]);
        rooms.insert(EasternPalace, &EASTERN_PALACE_ROOMS[..]);
        rooms.insert(DesertPalace, &DESERT_PALACE_ROOMS[..]);
        rooms.insert(TowerOfHera, &TOWER_OF_HERA_ROOMS[..]);
        rooms.insert(AgahnimsTower, &AGAHNIMS_TOWER_ROOMS[..]);
        rooms.insert(PalaceOfDarkness, &PALACE_OF_DARKNESS_ROOMS[..]);
        rooms.insert(SwampPalace, &SWAMP_PALACE_ROOMS[..]);
        rooms.insert(SkullWoods, &SKULL_WOODS_ROOMS[..]);
        rooms.insert(ThievesTown, &THIEVES_TOWN_ROOMS[..]);
        rooms.insert(IcePalace, &ICE_PALACE_ROOMS[..]);
        rooms.insert(MiseryMire, &MISERY_MIRE_ROOMS[..]);
        rooms.insert(TurtleRock, &TURTLE_ROCK_ROOMS[..]);
        rooms.insert(GanonsTower, &GANONS_TOWER_ROOMS[..]);

        rooms
    };
    pub static ref DUNGEON_KEYS: HashMap<Dungeon, ItemId> = {
        use Dungeon::*;
        use ItemId::*;

        let mut keys = HashMap::new();

        keys.insert(HyruleCastle, HyruleKey);
        keys.insert(EasternPalace, EasternKey);
        keys.insert(DesertPalace, DesertKey);
        keys.insert(TowerOfHera, HeraKey);
        keys.insert(AgahnimsTower, AgaKey);
        keys.insert(PalaceOfDarkness, PodKey);
        keys.insert(SwampPalace, SwampKey);
        keys.insert(SkullWoods, SkullKey);
        keys.insert(ThievesTown, ThievesKey);
        keys.insert(IcePalace, IceKey);
        keys.insert(MiseryMire, MireKey);
        keys.insert(TurtleRock, TurtleKey);
        keys.insert(GanonsTower, GTKey);

        keys
    };
    pub static ref DUNGEON_BIG_KEYS: HashMap<Dungeon, ItemId> = {
        use Dungeon::*;
        use ItemId::*;

        let mut keys = HashMap::new();

        keys.insert(HyruleCastle, HyruleBigKey);
        keys.insert(EasternPalace, EasternBigKey);
        keys.insert(DesertPalace, DesertBigKey);
        keys.insert(TowerOfHera, HeraBigKey);
        keys.insert(AgahnimsTower, AgaBigKey);
        keys.insert(PalaceOfDarkness, PodBigKey);
        keys.insert(SwampPalace, SwampBigKey);
        keys.insert(SkullWoods, SkullBigKey);
        keys.insert(ThievesTown, ThievesBigKey);
        keys.insert(IcePalace, IceBigKey);
        keys.insert(MiseryMire, MireBigKey);
        keys.insert(TurtleRock, TurtleBigKey);
        keys.insert(GanonsTower, GTBigKey);

        keys
    };
}
